/* Copyright (c) 2023 Gesellschaft für wissenschaftliche Datenverarbeitung mbH Göttingen
  
   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:
  
   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.
  
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
   THE SOFTWARE.
*/

#include <iostream>

#include "cublas_v2.h"


/* CUDA kernels to initialize the values in the arrays. */

__global__ void ssetvalues(int n, float *x, float *y)
{
    int start = blockDim.x * blockIdx.x + threadIdx.x;
    int stride = blockDim.x * gridDim.x;
    for (int i = start; i < n; i += stride)
    {
        x[i] = (float)i;
        y[i] = 0.1 * (float)(n - i);
    }
}

__global__ void dsetvalues(int n, double *x, double *y)
{
    int start = blockDim.x * blockIdx.x + threadIdx.x;
    int stride = blockDim.x * gridDim.x;
    for (int i = start; i < n; i += stride)
    {
        x[i] = (double)i;
        y[i] = 0.1 * (double)(n - i);
    }
}


/* Naive cuda implementations of saxpy and daxpy. Their performance is way worse
   than saxpy and daxpy from cublas. */

__global__ void naive_saxpy(int n, float a, float *x, float *y)
{
    int start = blockDim.x * blockIdx.x + threadIdx.x;
    int stride = blockDim.x * gridDim.x;
    for (int i = start; i < n; i += stride)
        y[i] = a * x[i] + y[i];
}

__global__ void naive_daxpy(int n, double a, double *x, double *y)
{
    int start = blockDim.x * blockIdx.x + threadIdx.x;
    int stride = blockDim.x * gridDim.x;
    for (int i = start; i < n; i += stride)
        y[i] = a * x[i] + y[i];
}




/* Everything will be done for both single precision (float) and double
   precision (double) to show both saxpy and daxpy. */

int main()
{
    // The number of elements to use in the arrays for axpy.
    int N = 1UL << 27;
    std::cout << "size: " << N << std::endl;

    /* The number of cuda threads we will use per block for the naive
       implementation of axpy. Should be a multiple of 32. */
    int threadsPerBlock = 32;

    /* The number of blocks must include all elements of N. If N is not
       divisible by the number of threads per block, the number of blocks
       must be rounded up. */
    int blocks = N / threadsPerBlock;
    if (threadsPerBlock * blocks < N)
        blocks++;

    // The size of the float and double arrays.
    size_t size_f = N * sizeof(float);
    size_t size_d = N * sizeof(double);

    // The a of axpy.
    float a_f = 1.3391;
    double a_d = 1.3391;

    // The arrays for x and y on the GPU.
    float *x_f = NULL;
    float *y_f = NULL;
    double *x_d = NULL;
    double *y_d = NULL;

    // Allocate the arrays and check for failure.
    if (cudaSuccess != cudaMalloc(&x_f, size_f)
        || cudaSuccess != cudaMalloc(&y_f, size_f)
        || cudaSuccess != cudaMalloc(&x_d, size_d)
        || cudaSuccess != cudaMalloc(&y_d, size_d))
    {
        std::cout << "cudaMalloc error!" << std::endl;
        if (x_f != NULL)
            cudaFree(x_f);
        if (y_f != NULL)
            cudaFree(y_f);
        if (x_d != NULL)
            cudaFree(x_d);
        if (y_d != NULL)
            cudaFree(y_d);
        return 1;
    }

    // Make two cuda events for benchmarking.
    cudaEvent_t start, stop;
    if (cudaSuccess != cudaEventCreate(&start))
    {
        std::cout << "cudaEventCreate error!" << std::endl;
        cudaFree(x_f);
        cudaFree(y_f);
        cudaFree(x_d);
        cudaFree(y_d);
        return 1;
    }
    if (cudaSuccess != cudaEventCreate(&stop))
    {
        std::cout << "cudaEventCreate error!" << std::endl;
        cudaFree(x_f);
        cudaFree(y_f);
        cudaFree(x_d);
        cudaFree(y_d);
        cudaEventDestroy(start);
        return 1;
    }

    // Make the cublas handle.
    cublasHandle_t handle;
    if (CUBLAS_STATUS_SUCCESS != cublasCreate(&handle))
    {
        std::cout << "cudablasCreate error!" << std::endl;
        cudaFree(x_f);
        cudaFree(y_f);
        cudaFree(x_d);
        cudaFree(y_d);
        cudaEventDestroy(start);
        cudaEventDestroy(stop);
        return 1;
    }

    
    /* For the benchmarking, we will use the events and run the kernel multiple
       times till the total elapsed time exceeds the desired amount while
       keeping track of how many times it has been run. The first execution will
       be ignored. Cuda events work in milliseconds.*/
    const float bench_period = 10.0e3;
    float millis_so_far;
    float millis_from_event;
    size_t count;

    /* Benchmark naive saxpy. Before each run, we must reset the values of x and
       y. */
    count = 0;
    millis_so_far = 0.0;
    while (millis_so_far < bench_period)
    {
        ssetvalues<<<blocks, threadsPerBlock>>>(N, x_f, y_f);
        cudaEventRecord(start);
        naive_saxpy<<<blocks, threadsPerBlock>>>(N, a_f, x_f, y_f);
        cudaEventRecord(stop);
        cudaEventSynchronize(stop);
        cudaEventElapsedTime(&millis_from_event, start, stop);
        if (count != 0)
            millis_so_far += millis_from_event;
        count++;
    }

    // Print the result in microseconds per iteration.
    std::cout << "naive saxpy: " << 1e3 * millis_so_far / (float)(count - 1) << " us" << std::endl;

    // Do the same for cublas saxpy.
    count = 0;
    millis_so_far = 0.0;
    while (millis_so_far < bench_period)
    {
        ssetvalues<<<blocks, threadsPerBlock>>>(N, x_f, y_f);
        cudaEventRecord(start);
        cublasSaxpy(handle, N, &a_f, x_f, 1, y_f, 1);
        cudaEventRecord(stop);
        cudaEventSynchronize(stop);
        cudaEventElapsedTime(&millis_from_event, start, stop);
        if (count != 0)
            millis_so_far += millis_from_event;
        count++;
    }
    std::cout << "cublas saxpy: " << 1e3 * millis_so_far / (float)(count - 1) << " us" << std::endl;

    // Do the same for naive daxpy
    count = 0;
    millis_so_far = 0.0;
    while (millis_so_far < bench_period)
    {
        dsetvalues<<<blocks, threadsPerBlock>>>(N, x_d, y_d);
        cudaEventRecord(start);
        naive_daxpy<<<blocks, threadsPerBlock>>>(N, a_d, x_d, y_d);
        cudaEventRecord(stop);
        cudaEventSynchronize(stop);
        cudaEventElapsedTime(&millis_from_event, start, stop);
        if (count != 0)
            millis_so_far += millis_from_event;
        count++;
    }
    std::cout << "naive daxpy: " << 1e3 * millis_so_far / (float)(count - 1) << " us" << std::endl;

    // Do the same for cublas daxpy.
    count = 0;
    millis_so_far = 0.0;
    while (millis_so_far < bench_period)
    {
        dsetvalues<<<blocks, threadsPerBlock>>>(N, x_d, y_d);
        cudaEventRecord(start);
        cublasDaxpy(handle, N, &a_d, x_d, 1, y_d, 1);
        cudaEventRecord(stop);
        cudaEventSynchronize(stop);
        cudaEventElapsedTime(&millis_from_event, start, stop);
        if (count != 0)
            millis_so_far += millis_from_event;
        count++;
    }
    std::cout << "cublas daxpy: " << 1e3 * millis_so_far / (float)(count - 1) << " us" << std::endl;

    // Cleanup
    cudaFree(x_f);
    cudaFree(y_f);
    cudaFree(x_d);
    cudaFree(y_d);
    cudaEventDestroy(start);
    cudaEventDestroy(stop);
    cublasDestroy(handle);

    return 0;
}
